# Measure the time it takes to execute the file
start_time_AQMS <- Sys.time()

############################################################
#                                                          #
#                 Loop Ordinary Kriging                    #
#                                                          #
############################################################

print("Loop Ordinary Kriging")

if(time_resolution=="hourly"){
  #LOOCV_total and model_raw_at_AQMS is created empty to fill it later with LOOCV and raw model data respectivelly
  model_raw_at_AQMS           <- matrix(NA,ncol=nrow(local_stations@data),nrow=0) #list(length=nrow(date_times)) #we will have a large list with a data.frame for each hour
  colnames(model_raw_at_AQMS) <- local_stations$ref_obs #we want a matrix with empty data and the name of stations
  
}else if(time_resolution=="daily"){
  #generating a data.frame wih the information about local stations:
  cols              <- colnames(df_obs_sub[,1:(ncol(df_obs_sub)-1)]) #same as: colnames(subset(df_obs_sub, select = -date))
  selection         <- local_stations@data$ref_obs %in% cols
  selected_stations <- local_stations[selection, ]
  new_df            <- data.frame(coords = selected_stations@coords, data = selected_stations@data)
  
  # prova tropecientosmil per saber quina es la dif:
  vars_local       <- as.character(local_stations@data$ref_obs)
  vars_df          <- names(df_obs_sub)[1:(length(names(df_obs_sub))-1)]
  vars_comunes     <- intersect(vars_local, vars_df)
  #num_vars_comunes <- length(vars_comunes)
  
  
  # Select only the stations that are also in vars_comunes
  local_stations <- subset(local_stations,with(local_stations@data, ref_obs %in% vars_comunes))
  
  #No entenc perque les dues linies aquestes no van, es com que la matriu no s'inicialitza, pero bueno funciona amb el data.frame
  #model_raw_at_AQMS            <- matrix(NA,ncol=length(vars_comunes),nrow=0) #list(length=nrow(date_times)) #we will have a large list with a data.frame for each hour
  #colnames(model_raw_at_AQMS)  <- vars_comunes #nope
  
  model_raw_at_AQMS           <- data.frame(matrix(NA, ncol = length(vars_comunes), nrow = 0))
  colnames(model_raw_at_AQMS) <- local_stations@data$ref_obs
  
  # -------
  # To know if we have an observated local station that its not in local stations data frame, we gonna create a new data frame with the names of the variables
  different_names <- setdiff(names(df_obs_sub),vars_comunes) # which names in df_obs_sub are not in vars_comunes, ojo! l'ordre sí importa
  # Print the different names
  print(different_names)
  
} else {
  print("Wrong time_resolution")
}


#starting OK and LOOCV:
for (hour in 1:length_loop) {
  print(paste0("loop OK number:",hour))
  df_model <- model[,,hour]
  
  #if all the model for a certain hour is nan we cannot apply UK
  if (!any(is.na(df_model))){
    
    # 1. Interpolate model values at obs locations
    #    1.1. Set spatialpointsdataframe
    #raw model:
    df_model                  <- data.frame(lat=lat_model,lon=lon_model,pollutant=sapply(df_model,as.matrix))
    coordinates(df_model)     <- ~lon+lat
    sp::proj4string(df_model) <- crs_latlon
    
    #config the coordinates in UTM
    if (compute_in_UTM) {
      df_model <- spTransform(df_model, sp::CRS(crs_utm))
    }
    
    #observations:
    df_obs                   <- data.frame(lat=local_stations@coords[,2],lon=local_stations@coords[,1],pollutant=NA,code=local_stations$ref_obs)
    coordinates(df_obs)      <- ~lon+lat
    sp::proj4string(df_obs)  <- crs_latlon
    
    #config the coordinates in UTM
    if (compute_in_UTM) {
      df_obs <- spTransform(df_obs, sp::CRS(crs_utm))
    }
    
    #if(any(is.na(df_obs@data)) | isTRUE(df_obs@data$pollutant<0)){
     # df_obs$pollutant<- ifelse(is.na(df_obs$pollutant)|df_obs$pollutant<0,0,df_obs$pollutant)
      #print(paste0("Negative or NA in the observations has been substituted by zero in the hour:", " ", hour))
    #}
    
    #    1.2 Compute Variogram
    set.seed(1) #we are picking random points in the next line, set.seed permits us to have same points in each loop
    pick_random_points        <- sample(1:nrow(df_model),1000)
    #any(is.na(df_model$pollutant)) #comprovació que hauria de donar falç... 
    pollutant.fit_ok          <- automap::autofitVariogram(formula=pollutant~1, input_data=df_model[pick_random_points,],model="Ste")
    # Hem posat model='Ste' pk la segona hora del 01/06/2021 peta el variograma, mirar bé perque no s'ajusa
    
    #Save the first variogram executed for a certain pollutant if we already don't have one saved:
     if (!file.exists(variogram_ok)) {
       saveRDS(pollutant.fit_ok,variogram_ok)
     }
    #    1.3 Compute ordinary Kriging
    
    #Del model interpolo cap a les coordenades de les obs 'OK'
    ok_gstat <- gstat(formula=pollutant~1, data = df_model, model = pollutant.fit_ok$var_model, nmax=100)
    
    #Interpolacio del model a les coordenades del conjunt
    OK       <- predict(object=ok_gstat, newdata=df_obs)
    
    #If the variogram has singular matrix then it produces NAN's, so in that case we take a good Variogram instead:
    if(any(is.na(OK@data$var1.pred))){
      print("WARNING: VARIOGRAM DIVERGED! Using a precalculated VARIOGRAM")
      pollutant.fit_ok=readRDS(variogram_ok)
      ok_gstat <- gstat(formula=pollutant~1, data = df_model, model = pollutant.fit_ok$var_model, nmax=100)
      
      #Interpolacio del model a les coordenades del conjunt
      OK <-predict(object=ok_gstat, newdata=df_obs)
    }
    
    model_raw_at_AQMS_h           <- matrix(data=OK$var1.pred, nrow=1)
    colnames(model_raw_at_AQMS_h) <- df_obs$code
    
    model_raw_at_AQMS <-  plyr::rbind.fill(as.data.frame(model_raw_at_AQMS), as.data.frame(model_raw_at_AQMS_h))
    
  }
  
  else{
    print(paste0("WARNING: The model is corrupted or non-existen in hour/day : ", hour))
    model_raw_at_AQMS_h           <- matrix(data=NA, nrow=1,ncol=nrow(local_stations@data))
    colnames(model_raw_at_AQMS_h) <- local_stations@data$ref_obs
    model_raw_at_AQMS             <- plyr::rbind.fill(as.data.frame(model_raw_at_AQMS),as.data.frame(model_raw_at_AQMS_h))
  }
}

model_raw_at_AQMS_df       <- as.data.frame(model_raw_at_AQMS)

if(time_resolution=="hourly"){
  model_raw_at_AQMS_df <- cbind(date=date_times$date, model_raw_at_AQMS_df)
  
}else{
  #model_raw_at_AQMS_df$dates <- df_obs_sub$date
  model_raw_at_AQMS_df <- cbind(date=df_obs_sub$date, model_raw_at_AQMS_df)
  
}


if (save_OK){
  saveRDS(object=model_raw_at_AQMS_df, file=paste0(save_result,"model_raw_at_AQMS_",file_format, "_",pollutant,"_df.rds"))
  if(time_resolution=="daily"){
    saveRDS(object=different_names, file=paste0(save_result,"different_local_station_",file_format, "_",pollutant,".rds"))
  }
  print("model_raw_at_AQMS saved")
} else {
  print("model_raw_at_AQMS have not been saved")
}


end_time_AQMS    <- Sys.time()
temps_total_AQMS <- end_time_AQMS - start_time_AQMS
print(paste0("Interpolation of the model to AQMS, execution time:",temps_total_AQMS))
